#!/usr/bin/env -S jq -Rrf

def parse: [
  scan("(?<r>[#+@!?]|\\[((?&r)*)\\|((?&r)*)])") |
  if .[1] != null then {cond: .[1] | parse, body: .[2] | parse}
  else .[0]
  end
];

def new_state: {
    stack: [],
    memory: {},
};

def ensure_height($height):
  .stack |= [range([$height - (. | length), 0] | max) | 0] + .
;

def gcd($a; $b): if $b == 0 then $a else gcd($b; $a % $b) end;
def lcm($a; $b): $a * $b / gcd($a; $b);
def lcm: reduce .[] as $n (1; lcm(.; $n));

def probpipe(x; y):
  [x | [y]] | (map(length) | lcm) as $l |
  .[] as $p | range($l / ($p | length)) | $p[]
;

def for($vals; filter):
  if ($vals | length) == 0 then .
  else {val: $vals[0], state: .} | filter | for(vals[1:]; filter)
  end
;

def eval($code):
  def do_loop($bound; $loop): probpipe(
    eval($loop.cond);
    ensure_height(1) | .stack[-1] as $cond | del(.stack[-1]) |
    if $bound != 0 and $cond != 0 then
      eval($loop.body) | do_loop($bound - 1; $loop)
    end
  );

  for($code; . as {$val, $state} | $state |
    if $val == "#" then
      .stack += [0]
    elif $val == "+" then
      ensure_height(1) | .stack[-1] += 1
    elif $val == "@" then
      ensure_height(1) | .stack[-1] = .memory[.stack[-1] | tostring]
    elif $val == "!" then
      ensure_height(2) |
      .memory[.stack[-1] | tostring] = .stack[-2] |
      del(.stack[-1, -2])
    elif $val == "?" then
      ensure_height(1) |
      range(.stack[-1] + 1) as $n | .stack[-1] = $n
    else
      ensure_height(1) |
      .stack[-1] as $bound | del(.stack[-1]) | do_loop($bound; $val)
    end
  )
;

def run:
  parse as $code | new_state | eval($code) |
  ensure_height(1).stack[-1] | tostring
;

def most_frequent:
  reduce .[] as $k ({}; .[$k] += 1) | to_entries | max_by(.value).key
;

[run] | most_frequent
